------------------------------------------------------
-- ReagentCost.lua
-- Version 1600.1
------------------------------------------------------

FRC_Config = { };
FRC_Config.Enabled = true;
FRC_Config.MinProfitRatio = 0;
FRC_Config.MinProfitMoney = nil;

FRC_ReagentLinks = { };

local MIN_SCANS = 35; -- times an item must be seen at auction to be considered a good sample (equates to 100% on our confidence scale)
local MIN_CONFIDENCE = 5; -- cutoff so we don't report items we have little data on as potentially profitable

function FRC_CraftFrame_SetSelection(id)
	FRC_Orig_CraftFrame_SetSelection(id);
	
	if ( not id ) then
		return;
	end
	local name, rank, maxRank = GetCraftDisplaySkillLine();
	if not (name) then
		return;
	end
	local craftName, craftSubSpellName, craftType, numAvailable, isExpanded, trainingPointCost, requiredLevel = GetCraftInfo(id);
	if ( trainingPointCost and trainingPointCost > 0 ) then
		return;
	end
	if ( craftType == "header" ) then
		return;
	end

	local costText;
	if (FRC_Config.Enabled) then
		local itemLink = GetCraftItemLink(id);
		if not (itemLink) then
			itemLink = craftName; -- Some enchanting formulae produce items (runed rods), most don't. Enchants don't link, items do.
		end
		local materialsTotal, confidenceScore = FRC_MaterialsCost(name, itemLink);
		costText = LIGHTYELLOW_FONT_COLOR_CODE.."(Total cost: " .. FRC_GetTextGSC(materialsTotal) ..FONT_COLOR_CODE_CLOSE..GRAY_FONT_COLOR_CODE.." Confidence: "..confidenceScore.."%"..FONT_COLOR_CODE_CLOSE..LIGHTYELLOW_FONT_COLOR_CODE..")"..FONT_COLOR_CODE_CLOSE;
	end
	
	CraftReagentLabel:SetText(SPELL_REAGENTS.." "..costText);
	CraftReagentLabel:Show();
end

function FRC_TradeSkillFrame_SetSelection(id)
	FRC_Orig_TradeSkillFrame_SetSelection(id);
	
	if ( not id ) then
		return;
	end
	local skillName, skillType, numAvailable, isExpanded = GetTradeSkillInfo(id);
	if ( skillType == "header" ) then
		return;
	end
	local skillLineName, skillLineRank, skillLineMaxRank = GetTradeSkillLine();
	
	local costText;
	if (FRC_Config.Enabled) then
		local materialsTotal, confidenceScore = FRC_MaterialsCost(skillLineName, GetTradeSkillItemLink(id));
		costText = LIGHTYELLOW_FONT_COLOR_CODE.."(Total cost: " .. FRC_GetTextGSC(materialsTotal) ..FONT_COLOR_CODE_CLOSE..GRAY_FONT_COLOR_CODE.." Confidence: "..confidenceScore.."%"..FONT_COLOR_CODE_CLOSE..LIGHTYELLOW_FONT_COLOR_CODE..")"..FONT_COLOR_CODE_CLOSE;
	end
	
	TradeSkillReagentLabel:SetText(SPELL_REAGENTS.." "..costText);
	TradeSkillReagentLabel:Show();
end

function FRC_OnLoad()
	
	this:RegisterEvent("CRAFT_SHOW");
	this:RegisterEvent("CRAFT_UPDATE");
	this:RegisterEvent("TRADE_SKILL_SHOW");
	this:RegisterEvent("TRADE_SKILL_UPDATE");

	-- Register Slash Commands
	SLASH_FRC1 = "/reagentcost";
	SLASH_FRC2 = "/rc";
	SlashCmdList["FRC"] = function(msg)
		FRC_ChatCommandHandler(msg);
	end
	
	-- Overrides for displaying info in CraftFrame / TradeSkillFrame
	FRC_Orig_CraftFrame_SetSelection = CraftFrame_SetSelection;
	CraftFrame_SetSelection = FRC_CraftFrame_SetSelection;
	FRC_Orig_TradeSkillFrame_SetSelection = TradeSkillFrame_SetSelection;
	TradeSkillFrame_SetSelection = FRC_TradeSkillFrame_SetSelection;
	
	GFWUtils.Print(GFW_COLOR_START.."Fizzwidget Reagent Cost initialized!"..FONT_COLOR_CODE_CLOSE);

end

function FRC_OnEvent(event)

	if ( event == "TRADE_SKILL_SHOW" or event == "TRADE_SKILL_UPDATE" ) then
	
		local skillLineName, skillLineRank, skillLineMaxRank = GetTradeSkillLine();
		if (FRC_ReagentLinks == nil) then
			FRC_ReagentLinks = { };
		end
		if (FRC_ReagentLinks[skillLineName] == nil) then
			FRC_ReagentLinks[skillLineName] = { };
		end

		for id = GetNumTradeSkills(), 1, -1 do 
			-- loop from the bottom up, since the reagents we make for compound items are usually below the recipes that need them
			local itemName, skillType, numAvailable, isExpanded = GetTradeSkillInfo(id);
			if ( skillType ~= "header" ) then				
				local itemLink = GetTradeSkillItemLink(id);
				FRC_ReagentLinks[skillLineName][itemLink] = { };
				for i=1, GetTradeSkillNumReagents(id), 1 do
					local link = GetTradeSkillReagentItemLink(id, i);
					local reagentName, reagentTexture, reagentCount, playerReagentCount = GetTradeSkillReagentInfo(id, i);

					table.insert(FRC_ReagentLinks[skillLineName][itemLink], {link=link, count=reagentCount});
				end
			end
		end
		return;
		
	elseif ( event == "CRAFT_SHOW" or event == "CRAFT_UPDATE" ) then
	
		-- This is used only for Enchanting
		local skillLineName, rank, maxRank = GetCraftDisplaySkillLine();
		if not (skillLineName) then
			return; -- Hunters' Beast Training also uses the CraftFrame, but doesn't have a SkillLine.
		end
		if (FRC_ReagentLinks == nil) then
			FRC_ReagentLinks = { };
		end
		if (FRC_ReagentLinks[skillLineName] == nil) then
			FRC_ReagentLinks[skillLineName] = { };
		end

		for id = GetNumCrafts(), 1, -1 do
			if ( craftType ~= "header" ) then
				craftName, craftSubSpellName, craftType, numAvailable, isExpanded, trainingPointCost, requiredLevel = GetCraftInfo(id);
				local itemLink = GetCraftItemLink(id);
				if not (itemLink) then
					itemLink = craftName; -- may be an item, may be a (currently unlikable) enchant
				end
				FRC_ReagentLinks[skillLineName][itemLink] = { };
				for i=1, GetCraftNumReagents(id), 1 do
					local link = GetCraftReagentItemLink(id, i);
					local reagentName, reagentTexture, reagentCount, playerReagentCount = GetCraftReagentInfo(id, i);

					table.insert(FRC_ReagentLinks[skillLineName][itemLink], {link=link, count=reagentCount});
				end
			end
		end
		return;
		
	end

end

function FRC_ChatCommandHandler(msg)

	-- Print Help
	if ( msg == "help" ) or ( msg == "" ) then
		GFWUtils.Print(GFW_COLOR_START.."Fizzwidget Reagent Cost:"..FONT_COLOR_CODE_CLOSE);
		GFWUtils.Print(GFW_COLOR_START.."/reagentcost (or /rc) <command>"..FONT_COLOR_CODE_CLOSE);
		GFWUtils.Print("- "..GFW_COLOR_START.."help"..FONT_COLOR_CODE_CLOSE.." - Print this helplist.");
		GFWUtils.Print("- "..GFW_COLOR_START.."status"..FONT_COLOR_CODE_CLOSE.." - Check current settings.");
		GFWUtils.Print("- "..GFW_COLOR_START.."reset"..FONT_COLOR_CODE_CLOSE.." - Reset to default settings.");
		GFWUtils.Print("- "..GFW_COLOR_START.."on|off"..FONT_COLOR_CODE_CLOSE.." - Toggle displaying info in tradeskill windows.");
		GFWUtils.Print("- "..GFW_COLOR_START.."report [<skillname>]"..FONT_COLOR_CODE_CLOSE.." - Output a list of the most profitable tradeskill items you can make. (Or only those produced through <skillname>.)");
		GFWUtils.Print("- "..GFW_COLOR_START.."minprofit <number>"..FONT_COLOR_CODE_CLOSE.." - When reporting, only show items whose estimated profit is <number> or greater. (In copper, so 1g == 10000.)");
		GFWUtils.Print("- "..GFW_COLOR_START.."minprofit <number>%"..FONT_COLOR_CODE_CLOSE.." - When reporting, only show items whose estimated profit exceeds its cost of materials by <number> percent or more.");
		return;
	end
	
	-- Check Status
	if ( msg == "status" ) then
		if (FRC_Config.Enabled) then
			GFWUtils.Print(GFW_COLOR_START.."Reagent Cost "..FONT_COLOR_CODE_CLOSE.."is"..GFW_COLOR_START.." displaying materials cost in tradeskill windows."..FONT_COLOR_CODE_CLOSE);
		else
			GFWUtils.Print(GFW_COLOR_START.."Reagent Cost "..FONT_COLOR_CODE_CLOSE.."is not"..GFW_COLOR_START.." displaying materials cost in tradeskill windows."..FONT_COLOR_CODE_CLOSE);
		end
		if (FRC_Config.MinProfitMoney == nil) then
			GFWUtils.Print(GFW_COLOR_START.."Reports will only include items whose estimated profit exceeds materials cost by "..FONT_COLOR_CODE_CLOSE..FRC_Config.MinProfitRatio.."%"..GFW_COLOR_START.." or more."..FONT_COLOR_CODE_CLOSE);
		else
			GFWUtils.Print(GFW_COLOR_START.."Reports will only include items whose estimated profit is "..FONT_COLOR_CODE_CLOSE..FRC_GetTextGSC(FRC_Config.MinProfitMoney)..GFW_COLOR_START.." or greater."..FONT_COLOR_CODE_CLOSE);
		end
		return;
	end

	-- Reset Variables
	if ( msg == "reset" ) then
		FRC_Config.Enabled = true;
		FRC_Config.MinProfitRatio = 0;
		FRC_Config.MinProfitMoney = nil;
		GFWUtils.Print(GFW_COLOR_START.."Reagent Cost configuration reset."..FONT_COLOR_CODE_CLOSE);
		FRC_ChatCommandHandler("status");
		return;
	end
	
	-- Turn trade info gathering on
	if ( msg == "on" ) then
		FRC_Config.Enabled = true;
		GFWUtils.Print(GFW_COLOR_START.."Reagent Cost "..FONT_COLOR_CODE_CLOSE.."is"..GFW_COLOR_START.." displaying materials cost in tradeskill windows."..FONT_COLOR_CODE_CLOSE);
		return;
	end
	
	-- Turn trade info gathering Off
	if ( msg == "off" ) then
		FRC_Config.Enabled = false;
		GFWUtils.Print(GFW_COLOR_START.."Reagent Cost "..FONT_COLOR_CODE_CLOSE.."is not"..GFW_COLOR_START.." displaying materials cost in tradeskill windows."..FONT_COLOR_CODE_CLOSE);
		return;
	end

	if ( string.sub(msg, 1, 9) == "minprofit" ) then

		local _, _, number, isPercent = string.find(msg, "minprofit (-*%d+)(%%*)");
		if (number == nil) then
			GFWUtils.Print("Usage: "..GFW_COLOR_START.."/rc minprofit <number>[%]"..FONT_COLOR_CODE_CLOSE);
			return;
		end
		if (isPercent == "%") then
			FRC_Config.MinProfitRatio = tonumber(number);
			FRC_Config.MinProfitMoney = nil;
			GFWUtils.Print(GFW_COLOR_START.."Reports will only include items whose estimated profit exceeds materials cost by "..FONT_COLOR_CODE_CLOSE..number.."%"..GFW_COLOR_START.." or more."..FONT_COLOR_CODE_CLOSE);
		else
			FRC_Config.MinProfitRatio = nil;
			FRC_Config.MinProfitMoney = tonumber(number);
			GFWUtils.Print(GFW_COLOR_START.."Reports will only include items whose estimated profit is "..FONT_COLOR_CODE_CLOSE..FRC_GetTextGSC(number)..GFW_COLOR_START.." or greater."..FONT_COLOR_CODE_CLOSE);
		end
		return;
	end

	if ( string.sub(msg, 1, 6) == "report" ) then
		
		-- parse skill names from args
		local mySkills = { };
		local args = string.sub(msg, 8);
		for word in string.gfind(args, "%w+") do
			table.insert(mySkills, word);
		end
		
		-- if no args, use the skills this character knows
		if (table.getn(mySkills) == 0) then
			for skillIndex = 1, GetNumSkillLines() do
				local skillName, header, isExpanded, skillRank, numTempPoints, skillModifier, skillMaxRank, isAbandonable, stepCost, rankCost, minLevel, skillCostType = GetSkillLineInfo(skillIndex);
				if (isAbandonable) then
					table.insert(mySkills, skillName);
				end
			end
		end
		
		for _, skillName in mySkills do
			local knownItems = 0;
			local reliableItems = 0;
			local shownItems = 0;
			local itemsTable = FRC_ReagentLinks[skillName];
			if (itemsTable == nil) then
				if (ReagentData['reverseprofessions'][skillName] ~= nil) then
					-- don't want to barf errors about gathering skills...
					GFWUtils.Print(GFW_COLOR_START.."ReagentCost doesn't have information on "..FONT_COLOR_CODE_CLOSE..skillName..GFW_COLOR_START..". Please open your "..FONT_COLOR_CODE_CLOSE..skillName..GFW_COLOR_START.." window before requesting a report."..FONT_COLOR_CODE_CLOSE);
				end
			else
				local reportTable = { }; -- separate report for each skill
				
				-- first, build a table that includes current Auctioneer prices for composite items
				for itemLink in itemsTable do
					knownItems = knownItems + 1;
					local itemPrice, itemConfidence = FRC_TypicalItemPrice(itemLink);
					local materialsCost, matsConfidence = FRC_MaterialsCost(skillName, itemLink);
					if (itemConfidence >= MIN_CONFIDENCE and matsConfidence >= MIN_CONFIDENCE) then
						reliableItems = reliableItems + 1;
						local profit = itemPrice - materialsCost;
						table.insert(reportTable, {link=itemLink, matsCost=materialsCost, matsConf=matsConfidence, itemPrice=itemPrice, itemConf=itemConfidence, profit=profit});
					end
				end
				
				if (table.getn(reportTable) == 0) then 
					GFWUtils.Print(GFW_COLOR_START.."ReagentCost doesn't have information on "..FONT_COLOR_CODE_CLOSE..skillName..GFW_COLOR_START..". Please open your "..FONT_COLOR_CODE_CLOSE..skillName..GFW_COLOR_START.." window before requesting a report."..FONT_COLOR_CODE_CLOSE);
				else
					GFWUtils.Print(GFW_COLOR_START.."Most profitable recipes for "..FONT_COLOR_CODE_CLOSE..skillName..":");

					-- then sort by profitability
					table.sort(reportTable, FRC_SortProfit);
					
					-- and report those that meet our minimum requirements
					for _, reportInfo in reportTable do
						if (FRC_Config.MinProfitRatio and (reportInfo.profit / reportInfo.matsCost * 100) >= FRC_Config.MinProfitRatio) then
							shownItems = shownItems + 1;
							FRC_PrintReportLine(reportInfo);
						elseif (FRC_Config.MinProfitMoney and reportInfo.profit >= FRC_Config.MinProfitMoney) then
							shownItems = shownItems + 1;
							FRC_PrintReportLine(reportInfo);
						end
					end
					GFWUtils.Print(knownItems..GFW_COLOR_START.." recipes known, "..FONT_COLOR_CODE_CLOSE..reliableItems..GFW_COLOR_START.." with auction data, "..FONT_COLOR_CODE_CLOSE..shownItems..GFW_COLOR_START.." above profit threshold."..FONT_COLOR_CODE_CLOSE);
				end	
			end
		end
		
		return;
	end

end

function FRC_SortProfit(a, b)
	-- sort by ratio or actual amount based on which we're using as cutoff
	if (FRC_Config.MinProfitRatio ~= nil) then
		return (a.profit / a.matsCost) > (b.profit / b.matsCost);
	else
		return a.profit > b.profit;
	end
end

function FRC_PrintReportLine(reportInfo)
	local reportLine = reportInfo.link..": ";
	reportLine = reportLine ..GFW_COLOR_START.."materials cost "..FONT_COLOR_CODE_CLOSE.. FRC_GetTextGSC(reportInfo.matsCost) ..GRAY_FONT_COLOR_CODE.." ("..reportInfo.matsConf.."%)"..FONT_COLOR_CODE_CLOSE..", "
	reportLine = reportLine ..GFW_COLOR_START.."auction price "..FONT_COLOR_CODE_CLOSE.. FRC_GetTextGSC(reportInfo.itemPrice) ..GRAY_FONT_COLOR_CODE.." ("..reportInfo.itemConf.."%)"..FONT_COLOR_CODE_CLOSE..", "
	if (reportInfo.profit >= 0) then
		reportLine = reportLine ..GFW_COLOR_START.."profit "..FONT_COLOR_CODE_CLOSE.. FRC_GetTextGSC(reportInfo.profit);
	else
		reportLine = reportLine ..RED_FONT_COLOR_CODE.."loss "..FONT_COLOR_CODE_CLOSE.. FRC_GetTextGSC(reportInfo.profit);
	end
	GFWUtils.Print(reportLine);
end

function FRC_MaterialsCost(skillName, itemLink)
	local materialsTotal = 0;
	local totalConfidence = 0;
	local numAuctionReagents = 0;
		
	if (FRC_ReagentLinks[skillName][itemLink] == nil) then
		return -1, -1;
	end
	
	for _, reagentInfo in FRC_ReagentLinks[skillName][itemLink] do
		local price, confidence;
		local itemPrice, itemConfidence = FRC_TypicalItemPrice(reagentInfo.link);
		local subReagentsPrice, subReagentsConfidence = FRC_MaterialsCost(skillName, reagentInfo.link);
		if (subReagentsConfidence > itemConfidence) then
			price = subReagentsPrice;
			confidence = subReagentsConfidence;
		else
			price = itemPrice;
			confidence = itemConfidence;
		end
		materialsTotal = materialsTotal + (price * reagentInfo.count);
		if (confidence >= 0) then
			totalConfidence = totalConfidence + confidence;
			numAuctionReagents = numAuctionReagents + 1;
		end		
	end
	local confidenceScore = math.floor((totalConfidence / numAuctionReagents) * 100) / 100;
	
	return materialsTotal, confidenceScore;

end


function FRC_TypicalItemPrice(itemLink)
	local itemID, randomProp, enchant, uniqID, name = Auctioneer_BreakLink(itemLink);
	local itemKey = itemID..":"..randomProp..":"..enchant;	
	local medianPrice, medianCount = Auctioneer_GetItemHistoricalMedianBuyout(itemKey, name);
	local vendorPrice = 0;
	if (Auctioneer_BasePrices[itemID]) then
		vendorPrice = Auctioneer_BasePrices[itemID].b;
	end
	
	if (ReagentData_IsVendorItem(name)) then
		return vendorPrice, -1;
	elseif (medianPrice == 0) then
		return vendorPrice * 3, 0; -- generally a good guess for auction price if we don't have real auction data
	else
		return medianPrice, math.floor((medianCount / MIN_SCANS) * 1000) / 10; 
	end
end


-- Thanks to Auctioneer for cash-formatting functions!

-- calculate the gold, silver, and copper values based the ammount of copper
function FRC_GetGSC(money)
	if (money == nil) then money = 0; end
	local g = math.floor(money / 10000);
	local s = math.floor((money - (g*10000)) / 100);
	local c = math.floor(money - (g*10000) - (s*100));
	return g,s,c;
end

-- formats money text by color for gold, silver, copper
function FRC_GetTextGSC(money)
    local GSC_GOLD="ffd100";
    local GSC_SILVER="e6e6e6";
    local GSC_COPPER="c8602c";
    local GSC_START="|cff%s%d|r";
    local GSC_PART=".|cff%s%02d|r";
    local GSC_NONE="|cffa0a0a0none|r";

	local g, s, c = FRC_GetGSC(money);

	local gsc = "";
	if (g > 0) then
		gsc = format(GSC_START, GSC_GOLD, g);     
		if (s > 0) then
			gsc = gsc..format(GSC_PART, GSC_SILVER, s);
		end
	elseif (s > 0) then
		gsc = format(GSC_START, GSC_SILVER, s);
		if (c > 0) then
			gsc = gsc..format(GSC_PART, GSC_COPPER, c);
		end
	elseif (c > 0) then
		gsc = gsc..format(GSC_START, GSC_COPPER, c);
	else
		gsc = GSC_NONE;
	end

	return gsc;
end
